# Copyright (c) 2015, Sergiusz Bazanski
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""An API to retrieve and set data in the Warsaw Hackerspce LDAP tree."""

import ldap

from flask import g

from webapp import mc, cache_enabled, app


def connect():
    c = ldap.initialize(app.config['LDAP_URI'])
    c.start_tls_s()
    c.simple_bind_s(app.config['LDAP_BIND_DN'],
                    app.config['LDAP_BIND_PASSWORD'])
    return c

if not app.config.get('DISABLE_LDAP'):
    @app.before_request
    def _setup_ldap():
        g.ldap = connect()

    @app.teardown_request
    def _destroy_ldap(exception=None):
        g.ldap.unbind_s()

def get_ldap_group_diff(members):
    active_members = filter(lambda m: m['judgement'], members)
    fatty = set([member['username'] for member in active_members if member['type'] in ['fatty', 'supporting']])
    starving = set([member['username'] for member in active_members if member['type'] in ['starving']])
   
    ldap_fatty = set(get_group_members(g.ldap, 'fatty'))
    ldap_starving = set(get_group_members(g.ldap, 'starving'))
    ldap_potato = set(get_group_members(g.ldap, 'potato'))

    result = {}
    result['fatty_to_remove'] = list(ldap_fatty - fatty)
    result['fatty_to_add'] = list(fatty - ldap_fatty)
    result['starving_to_remove'] = list(ldap_starving - starving)
    result['starving_to_add'] = list(starving - ldap_starving)
    if sum([len(result[k]) for k in result]) == 0:
        return None
    return result

# kinda clunky with all the member fetching, transforming the list in various ways and updating it again here, but it's a workaround for LDAP crashing on modify_s, no fucks given
def update_member_groups(c, changes):
    for group in changes:
        target_members = set(get_group_members(g.ldap, group))
        changed = False
        for op in changes[group]:
            for username in changes[group][op]:
                if op == 'add':
                    changed = True
                    target_members.add(username)
                elif op == 'remove':
                    changed = True
                    target_members.remove(username)
        if not changed:
            continue
        values = []
        for username in target_members:
            values.append('uid={},{}'.format(username.encode('utf-8'),app.config['LDAP_USER_BASE']))
        modlist = [(ldap.MOD_REPLACE,'uniqueMember',values)]
        #print group,modlist
        c.modify_s('cn={},{}'.format(group.encode('utf-8'),app.config['LDAP_GROUP_BASE']), modlist)

# keeping it here instead of git history because it's preferable to the other method, as long as LDAP stops crashing
def update_member_groups_fucked(c, changes):
    ops = {'add': ldap.MOD_ADD, 'remove': ldap.MOD_DELETE}
    for group in changes:
        modlist = []
        for op in changes[group]:
            values = []
            for username in changes[group][op]:
                values.append('uid={},{}'.format(username.encode('utf-8'),app.config['LDAP_USER_BASE']))
            if values:
                modlist.append((ops[op],'uniqueMember',values))
        print group, modlist
        c.modify_s('cn={},{}'.format(group.encode('utf-8'),app.config['LDAP_GROUP_BASE']), modlist)

def get_group_members(c, group):
    lfilter = '(&(cn={}){})'.format(group, app.config['LDAP_GROUP_FILTER'])
    data = c.search_s(app.config['LDAP_GROUP_BASE'], ldap.SCOPE_SUBTREE,
                      lfilter, tuple(['uniqueMember',]))
    
    members = []
    for dn, obj in data:
        for k, v in obj.iteritems():
            if k == "uniqueMember":
                for iv in v:
                    part,uid,index = ldap.dn.str2dn(iv)[0][0]
                    if not part == 'uid' or not index == 1:
                        raise ValueError("First part type {} or index {} seem wrong for DN {}".format(part,index,iv))
                    members.append(uid.decode('utf-8'))
    return members

def get_member_fields(c, member, fields):
    if isinstance(fields, str):
        fields = [fields,]
    fields_needed = set(fields)
    fields_out = {}
    if cache_enabled:
        for field in fields:
            field_cache = mc.get('kasownik-ldap-member-{}/{}'
                                 .format(member, field))
            if field_cache is not None:
                fields_out[field] = field_cache
                fields_needed.remove(field)

    member = member.replace('(', '').replace(')', '')
    lfilter = '(&(uid={}){})'.format(member, app.config['LDAP_USER_FILTER'])
    data = c.search_s(app.config['LDAP_USER_BASE'], ldap.SCOPE_SUBTREE,
                      lfilter, tuple(fields))
    for dn, obj in data:
        for k, v in obj.iteritems():
            v = v[0].decode('utf-8')
            if k in fields_needed:
                fields_out[k] = v
                if cache_enabled:
                    mc.set('kasownik-ldap-member-{}/{}'
                           .format(member, field), v)

    for k in fields_needed - set(fields_out.keys()):
        fields_out[k] = None

    return fields_out
